//****************************************************************************
//**
//**    SYS_MAIN.CPP
//**    System Main
//**
//**    Copyright (c) 1998 Chris Hargrove
//**    Notice: Usage of any code in this file is subject to the rules
//**            described in the LICENSE.TXT file included in this directory.
//**            Reading, compiling, or otherwise using this code constitutes
//**            automatic acceptance of the rules in said text file.
//**
//****************************************************************************
//============================================================================
//    IMPLEMENTATION HEADERS
//============================================================================
#define SYS_DEFS_INCLUDE_WIN32 // for SYSW_Argc, SYSW_Argv,
                               // and SYSW_MessageBox
#include "sys_defs.h"
#include "sys_main.h"

//============================================================================
//    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
//============================================================================
#define SAFEMALLOC_MARKER 0x2b3d1234
#define SAFEMALLOC_USEMARKER // enable malloc/free marker verify

//============================================================================
//    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
//============================================================================
//============================================================================
//    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
//============================================================================
//============================================================================
//    IMPLEMENTATION PRIVATE DATA
//============================================================================
static float sys_baseTime;
static float sys_lastTime;

static dword testReceiverHandle, quitReceiverHandle;

//============================================================================
//    INTERFACE DATA
//============================================================================
char sys_programPath[SYS_MAX_PATH];
float sys_curTime, sys_frameTime;

//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
//============================================================================
static void UpdateFrameTimes();
static void ResetFrameTimes();

//============================================================================
//    IMPLEMENTATION PRIVATE FUNCTIONS
//============================================================================
static void UpdateFrameTimes()
{
	sys_curTime = M_TimeFloat() - sys_baseTime;
	sys_frameTime = sys_curTime - sys_lastTime;
	sys_lastTime = sys_curTime;
}

static void ResetFrameTimes()
{
	sys_baseTime = M_TimeFloat();
	sys_lastTime = sys_curTime = sys_frameTime = 0.0;
}

//============================================================================
//    INTERFACE FUNCTIONS
//============================================================================

// The "Big Three" control functions

// initialization
void SYS_Init()
{
	XF_Init(); // initialize file system

	if (!M_CmdLineOption("noalert", 0)) // example of command line option use
	{
		SYS_Alertf("Hi!  I'm a small test of the zone memory subsystem.\n\n"
                   "I will consume a zone of memory 256 bytes at a time "
                   "until it runs out, then error out.  The default\nzone "
                   "size is 1 MB, but you can change it by using the -size "
                   "option followed by a size in MB.\nYou can also change the "
                   "allocation size from 256 to something else with the -chunk "
                   "option\nfollowed by a size in bytes.\n\n"
                   "Since zones are paged directly by Windows, try creating "
                   "some large ones and see how it performs\n"
                   "for you! (ex. -size 64 -chunk 4096)\n\n"
				   "BTW, you can run this program with the -noalert "
				   "option to skip this message.");
	}

	SYS_InitDebug();

	//-----------------------------------------
	// call our zone memory test
	
    {
        byte parm;
        dword zoneSize=0, chunkSize=0;
        dword zoneHandle;
    
        if (parm = M_CmdLineOption("size", 1))
            zoneSize = atoi(SYS_Argv(parm+1))*1024*1024;
        if (!zoneSize)
            zoneSize = 1*1024*1024;
        if (parm = M_CmdLineOption("chunk", 1))
            chunkSize = atoi(SYS_Argv(parm+1));
        if (!chunkSize)
            chunkSize = 256;
        
        zoneHandle = MEM_ZoneCreate(zoneSize);

        byte *ptr;
        while(ptr = (byte *)MEM_ZoneAlloc(chunkSize, zoneHandle))
        {            
            SYS_Debugf("%d bytes\n", MEM_ZoneGetCurrent(zoneHandle));
        }

        SYS_Quit(); // should never get here (will error out first)
    }
	
	//-----------------------------------------

	IN_Init(); // initialize user input
	
	ResetFrameTimes(); // initialize frame times from this point
}

// main loop
void SYS_Frame()
{
	UpdateFrameTimes();
	
	// process user input
	IN_Frame();
}

// de-initialization
void SYS_Shutdown()
{	
	SYS_ShutdownDebug();
	IN_Shutdown();
}

// alert notification
void SYS_Alertf(char* text, ...)
{
	static char textBuffer[2048];
	va_list args;
	va_start(args, text);
	vsprintf(textBuffer, text, args);
	va_end(args);
	SYSW_MessageBox(APPTITLE " Notification", MB_OK, textBuffer);
}

// initialize debugging window
void SYS_InitDebug()
{
	SYSW_InitDebug();
}

// shut down debugging window
void SYS_ShutdownDebug()
{
	SYSW_ShutdownDebug();
}

// debugging window output
void SYS_Debugf(char* text, ...)
{
	static char textBuffer[1024];
	va_list args;
	va_start(args, text);
	vsprintf(textBuffer, text, args);
	va_end(args);
	SYSW_DebugString(textBuffer);
}

// exit without error
void SYS_Quit()
{
	SYS_Shutdown();
	exit(0);
}

// exit with fatal error
void SYS_Errorf(char *text, ...)
{
	static char textBuffer[1024];
	va_list args;
	va_start(args, text);
	vsprintf(textBuffer, text, args);
	va_end(args);
	SYSW_MessageBox("Fatal Error", MB_OK, textBuffer);
	SYS_Shutdown();
	exit(1);
}

// command line argument functions, call windows equivalents
byte SYS_Argc()
{
	return(SYSW_Argc());
}

char* SYS_Argv(byte index)
{
	return(SYSW_Argv(index));
}

// safe malloc and free wrappers, with optional marker verify (4 byte cost)
void* SYS_SafeMalloc(int size)
{
#ifdef SAFEMALLOC_USEMARKER	
	size += 4;
#endif
	void *res = malloc(size);
	if (!res)
		SYS_Errorf("SYS_SafeMalloc: Out of memory on %d bytes", size);
#ifdef SAFEMALLOC_USEMARKER	
	*((long *)res) = SAFEMALLOC_MARKER;
	return((long *)res+1);
#else
	return(res);
#endif
}

void SYS_SafeFree(void **ptr)
{
	long *p;
	if (!(*ptr))
		SYS_Errorf("SYS_SafeFree: Attempt to free null pointer");
	p = (long *)*ptr;
#ifdef SAFEMALLOC_USEMARKER	
	p -= 1;
	if (*p != SAFEMALLOC_MARKER)
    {
		if (*p == (SAFEMALLOC_MARKER ^ 0xFFFFFFFF))
            SYS_Errorf("SYS_SafeFree: Attempt to free pointer twice");
        else
            SYS_Errorf("SYS_SafeFree: Attempt to free non-SafeMalloc'ed pointer");
    }
    *p = SAFEMALLOC_MARKER ^ 0xFFFFFFFF;
#endif
	free(p);
	*ptr = NULL;
}

//============================================================================
//    INTERFACE CLASS BODIES
//============================================================================

//****************************************************************************
//**
//**    END MODULE SYS_MAIN.CPP
//**
//****************************************************************************

